from abc import ABC, abstractmethod
import hashlib
from http import HTTPStatus
from http.cookies import SimpleCookie
import json
import logging
from typing import Dict, List
import warnings
import weakref

import esprima
import requests
import urllib3

TIMEOUT = 5

_LOGGER = logging.getLogger(__name__)

def _encode_luci_string(unencoded_string):
    """Encodes a string to be sent to a G3100 gateway in a "luci_" POST parameter."""
    md5_hash = hashlib.md5(unencoded_string.encode('ascii')).hexdigest()
    return hashlib.sha512(md5_hash.encode('ascii')).hexdigest()


class Gateway(ABC):
    def __init__(self):
        super().__init__()

        self.connected_devices = {}
        self.success_init = False

    @abstractmethod
    def check_auth(self) -> bool:
        """Attempts to authenticate with the device.

        Returns whether or not authentication succeeded.
        """
        return NotImplementedError()

    @abstractmethod
    def get_connected_devices(self) -> Dict[str, str]:
        """Gets the connected devices as a MAC address -> hostname map."""
        return NotImplementedError()


class Gateway1100(Gateway):
    def __init__(self, host, password, use_https=True):
        super().__init__()

        self.verify = False
        
        if use_https:
            self.scheme = 'https'
            urllib3.disable_warnings(urllib3.exceptions.InsecureRequestWarning)
        else:
            self.scheme = 'http'

        self.host = self.scheme + '://' + host
        self.password = password

        self.session = requests.Session()

    def get_connected_devices(self):
        devices_raw = self.session.get(self.host + '/api/devices', timeout=TIMEOUT, verify=self.verify)
        devices = json.loads(devices_raw.text)
        return {device['mac']: device['name'] for device in devices if device['status']}

    def check_auth(self):
        res = self.session.get(self.host + '/api/devices', timeout=TIMEOUT, verify=self.verify)
        if res.status_code == 200:
            return True

        getLogin = self.session.get(self.host + '/api/login', timeout=TIMEOUT, verify=self.verify)
        salt = getLogin.json()['passwordSalt']

        encodedPassword = hashlib.sha512()
        encodedPassword.update((self.password + salt).encode('ascii'))

        payload = json.dumps({"password": encodedPassword.hexdigest()})

        postLogin = self.session.post(self.host + '/api/login', data=payload, timeout=TIMEOUT, verify=self.verify)
        token = SimpleCookie(postLogin.headers.get('set-cookie'))['XSRF-TOKEN'].value

        self.session.headers.update({'X-XSRF-TOKEN': token})

        res = self.session.get(self.host + '/api/devices', timeout=TIMEOUT, verify=self.verify)
        if res.status_code == 200:
            return True
        return False


class Gateway3100(Gateway):
    def __init__(self, host, password):
        super().__init__()

        urllib3.disable_warnings(urllib3.exceptions.InsecureRequestWarning)

        self.verify = False
        self.host = 'https://' + host
        self.username = 'admin'
        self.password = password
        self.token = ''

        self.session = requests.Session()

        # Attempt to log out when this object is destroyed.
        weakref.finalize(
            self,
            self.session.post,
            self.host + '/logout.cgi',
            timeout=TIMEOUT,
            verify=self.verify,
            data={'token': self.token},
        )

    @classmethod
    def _is_valid_host(cls, host):
        urllib3.disable_warnings(urllib3.exceptions.InsecureRequestWarning)
        return requests.get('https://' + host + '/loginStatus.cgi', verify=False).status_code != HTTPStatus.NOT_FOUND

    def get_connected_devices(self):
        res = self.session.get(
            self.host + '/cgi/cgi_owl.js', timeout=TIMEOUT, verify=self.verify
        )

        if res.status_code != HTTPStatus.OK:
            _LOGGER.warning('Failed to get connected devices from gateway; '
                            'got HTTP status code %s', res.status_code)

        connected_devices = {}

        # Unfortunately, the data is provided to the frontend not as a JSON
        # blob, but as some JavaScript to execute.  The below code uses a
        # JavaScript parser and AST visitor to extract the known device data
        # from the script.
        #
        # Example response:
        #
        # addROD('known_device_list', { 'known_devices': [ { 'mac': 'xx:xx:xx:xx:xx:xx', 'hostname': 'name' } ] });
        def visitor(node, metadata):
            if node.type != 'CallExpression':
                return

            if node.callee.type != 'Identifier' or node.callee.name != 'addROD':
                return

            if node.arguments[0].value != 'known_device_list':
                return

            known_devices_node = None
            for prop in node.arguments[1].properties:
                if prop.key.value == 'known_devices':
                    known_devices_node = prop.value

            if known_devices_node is None:
                _LOGGER.debug('Failed to find known_devices object in response data')
                return

            for device in known_devices_node.elements:
                data = {prop.key.value: prop.value.value for prop in device.properties}
                if 'activity' not in data or 'mac' not in data or 'hostname' not in data:
                    continue
                if data['activity'] == 1:
                    connected_devices[data['mac']] = data['hostname']

        esprima.parseScript(res.text, {}, visitor)

        return connected_devices

    def _check_login_status(self):
        res = self.session.get(
            self.host + '/loginStatus.cgi', timeout=TIMEOUT, verify=self.verify
        )
        if res.status_code == HTTPStatus.OK and res.json()['islogin'] == '1':
            # Store the XSRF token for use in future requests.
            self.token = res.json()['token']
            return True
        return False

    def check_auth(self):
        if self._check_login_status():
            return True

        body = {
            'luci_username': _encode_luci_string(self.username),
            'luci_password': _encode_luci_string(self.password),
        }
        res = self.session.post(
            self.host + '/login.cgi', timeout=TIMEOUT, verify=self.verify, data=body
        )

        if res.status_code in (HTTPStatus.OK, HTTPStatus.FOUND):
            return self._check_login_status()

        if res.status_code == HTTPStatus.FORBIDDEN:
            response_json = res.json()
            if response_json.get('flag') == 2:
                _LOGGER.warning('Hit maximum session limit of %s sessions',
                                response_json['maxsession'])
        else:
            _LOGGER.debug('unexpected response code: %s', res.status_code)

        return False


class QuantumGatewayScanner:
    def __init__(self, host, password, use_https=True):
        self._gateway = self._get_gateway(host, password, use_https)
        self.success_init = self._gateway.check_auth()

    def _get_gateway(self, host, password, use_https) -> Gateway:
        if Gateway3100._is_valid_host(host):
            return Gateway3100(host, password)
        else:
            return Gateway1100(host, password, use_https)

    def scan_devices(self) -> List[str]:
        self.connected_devices = {}
        if self._gateway.check_auth():
            self.connected_devices = self._gateway.get_connected_devices()
        return self.connected_devices.keys()

    def get_device_name(self, device: str) -> str:
        return self.connected_devices.get(device)


from http import HTTPStatus
import unittest
from unittest import mock
import json
import hashlib

from requests import status_codes
import requests
from requests.api import request
import requests_mock
import re

from quantum_gateway import Gateway, Gateway1100, Gateway3100, QuantumGatewayScanner

G3100_LOGIN_STATUS_MATCHER = re.compile('^.*/loginStatus.cgi$')


class TestScanner(unittest.TestCase):

    @requests_mock.Mocker()
    def test_detects_g3100(self, m):
        m.get(G3100_LOGIN_STATUS_MATCHER)
        with mock.patch('quantum_gateway.Gateway3100') as MockGateway3100:
            scanner = QuantumGatewayScanner("192.168.1.1", "password")
            MockGateway3100.assert_called_once_with("192.168.1.1", "password")

    @requests_mock.Mocker()
    def test_defaults_to_g1100(self, m):
        m.get(G3100_LOGIN_STATUS_MATCHER, status_code=HTTPStatus.NOT_FOUND)
        with mock.patch('quantum_gateway.Gateway1100') as MockGateway1100:
            scanner = QuantumGatewayScanner("192.168.1.1", "password")
            MockGateway1100.assert_called_once_with("192.168.1.1", "password", True)

    def test_successful_init(self):
        with mock.patch.object(QuantumGatewayScanner, "_get_gateway") as mock_get_gateway:
            mock_gateway = mock.create_autospec(Gateway)
            mock_gateway.check_auth.return_value = True
            mock_get_gateway.return_value = mock_gateway

            scanner = QuantumGatewayScanner("192.168.1.1", "password")
            self.assertTrue(scanner.success_init)

    def test_failed_init(self):
        with mock.patch.object(QuantumGatewayScanner, "_get_gateway") as mock_get_gateway:
            mock_gateway = mock.create_autospec(Gateway)
            mock_gateway.check_auth.return_value = False
            mock_get_gateway.return_value = mock_gateway

            scanner = QuantumGatewayScanner("192.168.1.1", "password")
            mock_gateway.check_auth.assert_called_once()
            self.assertFalse(scanner.success_init)

    def test_get_connected_devices(self):
        with mock.patch.object(QuantumGatewayScanner, "_get_gateway") as mock_get_gateway:
            mock_gateway = mock.create_autospec(Gateway)
            mock_gateway.check_auth.return_value = True
            mock_gateway.get_connected_devices.return_value = {"mac_address": "hostname"}
            mock_get_gateway.return_value = mock_gateway

            scanner = QuantumGatewayScanner("192.168.1.1", "password")
            self.assertCountEqual(scanner.scan_devices(), ["mac_address"])

    def test_get_device_name(self):
        with mock.patch.object(QuantumGatewayScanner, "_get_gateway") as mock_get_gateway:
            mock_gateway = mock.create_autospec(Gateway)
            mock_gateway.check_auth.return_value = True
            mock_gateway.get_connected_devices.return_value = {"mac_address": "hostname"}
            mock_get_gateway.return_value = mock_gateway

            scanner = QuantumGatewayScanner("192.168.1.1", "password")
            scanner.scan_devices()
            self.assertEqual(scanner.get_device_name("mac_address"), "hostname")


@requests_mock.Mocker()
class TestGateway1100(unittest.TestCase):
    DEVICES_MATCHER = re.compile('^.*/api/devices$')
    LOGIN_MATCHER = re.compile('^.*/api/login$')
    TOKEN = 'TEST_TOKEN'
    PASSWORD_SALT = 'TEST_SALT'
    CORRECT_PASSWORD = 'correct'
    WRONG_PASSWORD = 'wrong'
    CONNECTED_DEVICES = {'00:11:22:33:44:55': 'iphone', '00:00:00:00:00:00': 'computer'}
    SERVER_CONNECTED_DEVICES_RESPONSE = '[{"mac": "00:11:22:33:44:55", "name": "iphone", "status": true}, {"mac": "00:00:00:00:00:00", "name": "computer", "status": true}, {"mac": "11:11:11:11:11:11", "name": "disconnected", "status": false}]'

    logged_in = False

    def setUp(self):
        self.logged_in = False

    def test_login_success(self, m):
        self.setup_matcher(m)

        host = '192.168.1.2'
        password = self.CORRECT_PASSWORD
        gateway = Gateway1100(host, password)

        self.assertTrue(gateway.check_auth())

    def test_login_fail(self, m):
        self.setup_matcher(m)

        host = '192.100.100.5'
        password = self.WRONG_PASSWORD
        gateway = Gateway1100(host, password)

        self.assertFalse(gateway.check_auth())

    def test_get_connected_devices(self, m):
        self.setup_matcher(m)

        host = 'mywifigateway.com'
        password = self.CORRECT_PASSWORD

        gateway = Gateway1100(host, password)

        gateway.check_auth()
        self.assertEqual(gateway.get_connected_devices(), self.CONNECTED_DEVICES)

    def setup_matcher(self, m):
        def devices_callback(request, context):
            if self.is_logged_in(request):
                context.status_code = 200
                return self.SERVER_CONNECTED_DEVICES_RESPONSE
            else:
                context.status_code = 401

        def password_callback(request, context):
            if self.is_correct_password(request):
                context.status_code = 200
                self.logged_in = True
            else:
                context.status_code = 401
            context.headers['set-cookie'] = 'XSRF-TOKEN=' + self.TOKEN

        m.get(self.DEVICES_MATCHER, text=devices_callback)
        m.get(self.LOGIN_MATCHER, status_code=200, json={'passwordSalt': self.PASSWORD_SALT})
        m.post(self.LOGIN_MATCHER, text=password_callback)

    def is_logged_in(self, request):
        return self.logged_in and request.headers.get('X-XSRF-TOKEN') == self.TOKEN

    def is_correct_password(self, request):
        hash = hashlib.sha512()
        hash.update((self.CORRECT_PASSWORD + self.PASSWORD_SALT).encode('ascii'))
        expected_encoded_password = hash.hexdigest()
        actual_encoded_password = json.loads(request.body)['password']

        return actual_encoded_password == expected_encoded_password

@requests_mock.Mocker()
class TestGateway3100(unittest.TestCase):

    LOGIN_MATCHER = re.compile('^.*/login.cgi$')
    DEVICE_INFO_MATCHER = re.compile('^.*/cgi_owl.js$')
    LOGOUT_MATCHER = re.compile('^.*/logout.cgi$')

    LOGGED_OUT_STATUS_JSON = {"islogin": "0"}
    LOGGED_IN_STATUS_JSON = {"islogin": "1", "token": "token_value"}

    def setUp(self):
        self.host = "192.168.23.16"
        self.password = "password"

    def test_login_success(self, m):
        self._match_successful_login(m)
        self._match_logout(m)

        gateway = Gateway3100(self.host, self.password)

        self.assertTrue(gateway.check_auth())

    def test_login_fail(self, m):
        m.get(G3100_LOGIN_STATUS_MATCHER, json=self.LOGGED_OUT_STATUS_JSON)
        m.post(self.LOGIN_MATCHER, status_code=HTTPStatus.FORBIDDEN, json={})
        self._match_logout(m)

        gateway = Gateway3100(self.host, self.password)

        self.assertFalse(gateway.check_auth())

    def test_get_connected_devices(self, m):
        device_info_response_text = """
        addROD("known_device_list", { "known_devices": [
            { "mac": "xx:xx:xx:xx:xx:xx", "hostname": "active_device", "activity": 1 },
            { "mac": "xx:xx:xx:xx:xx:ab", "hostname": "inactive_device", "activity": 0 }
        ] });
        """

        self._match_successful_login(m)
        m.get(self.DEVICE_INFO_MATCHER, text=device_info_response_text)
        self._match_logout(m)

        gateway = Gateway3100(self.host, self.password)

        self.assertCountEqual(gateway.get_connected_devices(), {"xx:xx:xx:xx:xx:xx": "hostname"})

    def _match_successful_login(self, m):
        m.get(G3100_LOGIN_STATUS_MATCHER, json=self.LOGGED_OUT_STATUS_JSON)
        def check_login_data(request, context):
            self.assertEqual(request.data, {
                "luci_username": "",
                "luci_password": "",
            })
        m.post(self.LOGIN_MATCHER, text=check_login_data)
        m.get(G3100_LOGIN_STATUS_MATCHER, json=self.LOGGED_IN_STATUS_JSON)

    def _match_logout(self, m):
        m.post(self.LOGOUT_MATCHER)

if __name__ == '__main__':
    unittest.main()



from setuptools import setup

with open('README.md', 'r') as fh:
    long_description = fh.read()

with open('requirements.txt', 'r') as fh:
    requirements = fh.read().splitlines()

setup(
    name='quantum_gateway',
    version='0.0.6',
    author='Colby Rome',
    author_email='colbyrome@gmail.com',
    description='Query a Quantum Gateway',
    py_modules=['quantum_gateway'],
    install_requires=requirements,
    package_dir={'': 'src'},
    url='https://github.com/cisasteelersfan/quantum_gateway',
    long_description=long_description,
    long_description_content_type="text/markdown",
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
)
